package sample;

import java.lang.invoke.*;

public class Sample {
    public void tick() {
        System.out.println("tick");
    }

    public void tock() {
        System.out.println("tock");
    }

    static MethodHandle tickHandle;
    static MethodHandle tockHandle;

    public static void main(String[] args) throws Throwable {
        CallSite callSite = createHandle();
        MethodHandle methodHandle = callSite.dynamicInvoker();

        final int MAX = 10;
        for (int i = 0; i < MAX; i++) {
            methodHandle.invoke();

            Thread.sleep(1000);
            if(callSite.getTarget() == tickHandle)
                callSite.setTarget(tockHandle);
            else
                callSite.setTarget(tickHandle);
            if(i == MAX - 2)
                callSite.setTarget(createHandleToABlock());
        }
    }

    private static MethodHandle createHandleToABlock() throws NoSuchMethodException, IllegalAccessException {
        Runnable block = () -> System.out.println("tada");
        MethodHandles.Lookup lookup = MethodHandles.lookup();
        return lookup.findVirtual(block.getClass(), "run", MethodType.methodType(void.class)).bindTo(block);
    }

    private static CallSite createHandle() throws NoSuchMethodException, IllegalAccessException {
        MethodHandles.Lookup lookup = MethodHandles.lookup();
        Sample sample = new Sample();
        tickHandle = lookup.findVirtual(Sample.class, "tick", MethodType.methodType(void.class))
                .bindTo(sample);
        tockHandle = lookup.findVirtual(Sample.class, "tock", MethodType.methodType(void.class))
                .bindTo(sample);

        return new MutableCallSite(tickHandle);
    }
}
