package sample;

import java.util.Arrays;
import java.util.List;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.function.ToIntFunction;

//Run javap -c Sample on the bytecode to look at invokedynamic being used under the covers.
public class Sample {
    public static void main(String[] args) {
        List<Integer> numbers = Arrays.asList(1, 2, 3, 4, 5, 6);
        final int MAX = 1000_000;

        timeIt(MAX, () -> totalUsingAnonymousInnerClass(numbers));//Time: 0.347313
        timeIt(MAX, () -> totalUsingLambdaExpressions(numbers)); //Time: 0.275343
    }

    private static void totalUsingAnonymousInnerClass(final List<Integer> numbers) {
        numbers.stream().map(new Function<Integer, Integer>() {
            @Override
            public Integer apply(Integer value) {
                return value * 2;
            }
        }).reduce(0, new BinaryOperator<Integer>() {
            @Override
            public Integer apply(Integer value1, Integer value2) {
                return value1 + value2;
            }
        });
    }

    private static void totalUsingLambdaExpressions(final List<Integer> numbers) {
        ToIntFunction<Integer> toIntFunction = value -> value * 2;
        numbers.stream().map(toIntFunction).reduce(0, (value1, value2) -> value1 + value2);
    }

    private static void totalImperative(final List<Integer> numbers) {
        int total = 0;
        for(int element : numbers) {
            total += element * 2;
        }
    }

    private static void timeIt(final int times, Runnable runnable) {
        final long start = System.nanoTime();
        for(int i = 0; i < times; i++) {
            runnable.run();
        }
        System.out.println("Time: " + (System.nanoTime() - start)/1.0e9);
    }
}